@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004, 2005, 2007-2010, 2013  Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (C) 2004, 2005, 2007-2010, 2013, 2014, 2016  Internet Systems Consortium, Inc. ("ISC")
  * Copyright (C) 1999-2003  Internet Software Consortium.
  *
  * Permission to use, copy, modify, and/or distribute this software for any
@@ -31,7 +31,11 @@
 #include <time.h>
 #include <unistd.h>
 
+#ifndef WIN32
 #include <sys/wait.h>
+#else
+#include <direct.h>
+#endif
 
 #include <isc/boolean.h>
 #include <isc/commandline.h>
@@ -84,6 +88,9 @@ static char		T_tvec[T_MAXTESTS / 8];
 static char *		T_env[T_MAXENV + 1];
 static char		T_buf[T_BIGBUF];
 static char *		T_dir;
+#ifdef WIN32
+static testspec_t	T_testlist[T_MAXTESTS];
+#endif
 
 static int
 t_initconf(const char *path);
@@ -111,20 +118,31 @@ t_sighandler(int sig) {
 }
 
 int
-main(int argc, char **argv) {
+#ifndef WIN32
+main(int argc, char **argv)
+#else
+t_main(int argc, char **argv)
+#endif
+{
 	int			c;
 	int			tnum;
+#ifndef WIN32
 	int			subprocs;
 	pid_t			deadpid;
 	int			status;
+#endif
 	int			len;
 	isc_boolean_t		first;
 	testspec_t		*pts;
+#ifndef WIN32
 	struct sigaction	sa;
+#endif
 
 	isc_mem_debugging = ISC_MEM_DEBUGRECORD;
 	first = ISC_TRUE;
+#ifndef WIN32
 	subprocs = 1;
+#endif
 	T_timeout = T_TCTOUT;
 
 	/*
@@ -203,7 +221,9 @@ main(int argc, char **argv) {
 			exit(0);
 		}
 		else if (c == 'x') {
+#ifndef WIN32
 			subprocs = 0;
+#endif
 		}
 		else if (c == 'q') {
 			T_timeout = atoi(isc_commandline_argument);
@@ -240,12 +260,14 @@ main(int argc, char **argv) {
 	 * Setup signals.
 	 */
 
+#ifndef WIN32
 	sa.sa_flags = 0;
 	sigfillset(&sa.sa_mask);
 
 	sa.sa_handler = t_sighandler;
 	(void)sigaction(SIGINT,  &sa, NULL);
 	(void)sigaction(SIGALRM, &sa, NULL);
+#endif
 
 	/*
 	 * Output start stanza to journal.
@@ -273,8 +295,9 @@ main(int argc, char **argv) {
 
 	tnum = 0;
 	pts = &T_testlist[0];
-	while (*pts->pfv != NULL) {
+	while (pts->pfv != NULL) {
 		if (T_tvec[tnum / 8] & (0x01 << (tnum % 8))) {
+#ifndef WIN32
 			if (subprocs) {
 				T_pid = fork();
 				if (T_pid == 0) {
@@ -327,6 +350,9 @@ main(int argc, char **argv) {
 			else {
 				(*pts->pfv)();
 			}
+#else
+			(*pts->pfv)();
+#endif
 		}
 		++pts;
 		++tnum;
@@ -384,8 +410,8 @@ t_result(int result) {
 		case T_UNRESOLVED:
 			p = "UNRESOLVED";
 			break;
-		case T_UNSUPPORTED:
-			p = "UNSUPPORTED";
+		case T_SKIPPED:
+			p = "SKIPPED";
 			break;
 		case T_UNTESTED:
 			p = "UNTESTED";
@@ -525,7 +551,7 @@ t_fgetbs(FILE *fp) {
 		p = buf;
 		while ((c = fgetc(fp)) != EOF) {
 
-			if (c == '\n')
+			if ((c == '\r') || (c == '\n'))
 				break;
 
 			*p++ = c;
@@ -589,7 +615,7 @@ t_getdate(char *buf, size_t buflen) {
 /*
  * Some generally used utilities.
  */
-struct dns_errormap {
+static const struct dns_errormap {
 	isc_result_t	result;
 	const char *text;
 } dns_errormap[] = {
@@ -643,8 +669,8 @@ struct dns_errormap {
 isc_result_t
 t_dns_result_fromtext(char *name) {
 
-	isc_result_t		result;
-	struct dns_errormap	*pmap;
+	isc_result_t			result;
+	const struct dns_errormap	*pmap;
 
 	result = ISC_R_UNEXPECTED;
 
@@ -661,7 +687,7 @@ t_dns_result_fromtext(char *name) {
 	return (result);
 }
 
-struct dc_method_map {
+static const struct dc_method_map {
 	unsigned int	dc_method;
 	const char 	*text;
 } dc_method_map[] = {
@@ -674,8 +700,8 @@ struct dc_method_map {
 
 unsigned int
 t_dc_method_fromtext(char *name) {
-	unsigned int		dc_method;
-	struct dc_method_map	*pmap;
+	unsigned int			dc_method;
+	const struct dc_method_map	*pmap;
 
 	dc_method = DNS_COMPRESS_NONE;
 
@@ -736,11 +762,13 @@ t_eval(const char *filename, int (*func)(char **), int nargs) {
 	int		line;
 	int		cnt;
 	int		result;
+	int		tresult;
 	int		nfails;
 	int		nprobs;
 	int		npass;
 	char		*tokens[T_MAXTOKS + 1];
 
+	tresult = T_UNTESTED;
 	npass = 0;
 	nfails = 0;
 	nprobs = 0;
@@ -762,14 +790,15 @@ t_eval(const char *filename, int (*func)(char **), int nargs) {
 
 			cnt = t_bustline(p, tokens);
 			if (cnt == nargs) {
-				result = func(tokens);
-				switch (result) {
+				tresult = func(tokens);
+				switch (tresult) {
 				case T_PASS:
 					++npass;
 					break;
 				case T_FAIL:
 					++nfails;
 					break;
+				case T_SKIPPED:
 				case T_UNTESTED:
 					break;
 				default:
@@ -797,7 +826,24 @@ t_eval(const char *filename, int (*func)(char **), int nargs) {
 	else if (nfails > 0)
 		result = T_FAIL;
 	else if (npass == 0)
-		result = T_UNTESTED;
+		result = tresult;
 
 	return (result);
 }
+
+#ifdef WIN32
+void
+t_settests(const testspec_t list[]) {
+	int			tnum;
+	const testspec_t	*pts;
+
+	memset(T_testlist, 0, sizeof(T_testlist));
+
+	pts = &list[0];
+	for (tnum = 0; tnum < T_MAXTESTS - 1; pts++, tnum++) {
+		if (pts->pfv == NULL)
+			break;
+		T_testlist[tnum] = *pts;
+	}
+}
+#endif
